/*
Applet for displaying PIF files, with extension for GRN interaction dynamics
Copyright (C) 2008 Johannes Knabe
j.f.knabe@herts.ac.uk

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
import javax.swing.*;
import java.util.*;
import java.util.zip.*;
import java.net.URL;
import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;
import javax.swing.tree.*;
import javax.swing.event.*;
import javax.swing.filechooser.FileFilter;
import java.io.*;
import grph.*;

//reads files in the background, especially useful for online applets
class lazyReader extends Thread {

public grid2D grid;
public URL baseDir;
public File baseFile;
public Label loadprogress;

lazyReader(URL baseDir, grid2D grid){
  this.grid=grid;
  this.baseDir=baseDir;
}

lazyReader(File baseFile, grid2D grid){
  this.grid=grid;
  this.baseFile=baseFile;
}

public boolean readFile(int i){
  Pixel[][][] data;
  if(baseDir!=null){
    if(i>=grid.gridBuffer.size())
      data=grid.readPIF(null,i);
    else 
      data=grid.readPIF(baseDir+File.separator+grid.dirlist.get(i),i);
  }
  else try{
    data=grid.readPIF(baseFile+File.separator+grid.dirlist.get(i),i);
    }catch(ArrayIndexOutOfBoundsException e){data=null;}
  if(data==null){
    loadprogress.setVisible(false);
    return false;
  }
  loadprogress.setText("loaded: "+(i+1));
  synchronized(grid.gridBuffer){
    grid.gridBuffer.add(data);
  }
  return true;
}

public void run(){
  loadprogress.setVisible(true);
  int i=0;
  while(readFile(i))
    i++;
  grid.loadingFinished=true;
/*
  for(int i=0;i<grid.dirlist.size();i++)
    if(i>=grid.gridBuffer.size())
      readFile(i);
*/
}

}

class Pixel{

  int id;
  int type;
  double conc1;
  double conc2;

  Pixel(){type=0;id=-1; conc1=.0; conc2=.0;}
  Pixel(int id2,int type2,double conc12,double conc22){type=type2;id=id2; conc1=conc12; conc2=conc22;}

  public Pixel clone(){
  Pixel ret=new Pixel(id,type,conc1,conc2);
  return ret;
  }
}

class grid2D extends Canvas{

public static final int width=600; //fix width
public static final int height=400; //fix width
public static final int maxWide=600;   //and max Width
public static final int maxHigh=400;   //and max height

//you also might want to change: mapType2Col and mapConc2Col
public static int sizeX=60;
public static int sizeY=40;
public static int sizeZ=1;  //use "1" for 2D!
public static int whichZ=0; //obviously we can only draw 2D, so we have to decide on which "z" plane we want
public static int mode=0;

public static int psXstart=0;
public static int psYstart=0;
public static int psXresolution;
public static int psYresolution;

public static Vector<Double> minConcs1=new Vector<Double>();
public static Vector<Double> minConcs2=new Vector<Double>();
public static Vector<Double> maxConcs1=new Vector<Double>();
public static Vector<Double> maxConcs2=new Vector<Double>();

public static Color trackerColor=new Color(255,0,255);

public static ZipInputStream screens=null;
public static URL screensURL;

public static void resetScreens(){
 try{
  if(screens!=null) screens.close();
  screens=new ZipInputStream(screensURL.openStream());
 }catch(Exception e){System.err.println("Error reading Zipfile 2: "+screensURL);}
}

public static Pixel[][][] readPIF(String filename,int timer){
File file = null;
BufferedReader pif =null;
if(screens!=null){
 try{
  String fname="";
  if(filename!=null)
    fname=filename.substring(filename.indexOf(".zip")+5);//remove zipfile name
  ZipEntry t=null;
  int resetCount=0;
  while((t==null)||((!t.getName().equals(fname)))){
  	t=screens.getNextEntry();
  	if(t==null){
		if(filename==null){
			return null; //indicate that end of file has been reached, reset would have to be called externally now!
			}
		else{
			resetScreens();
			resetCount++;
			if(resetCount>1){System.err.println("ERROR: Could not find "+fname+", exiting"); System.exit(1);}
			}
		}
	else
		if(!t.getName().endsWith(".pif"))
			continue;
		else
			if(filename==null){
				String x=t.getName();
				if(x.endsWith(".pif"))
					dirlist.add(x);
				break;
			}
  }
  pif=new BufferedReader(new InputStreamReader(screens));
 }catch(Exception e){System.err.println("ERROR: ZipEntry "+filename+" found but unreadable, exiting! "+e); System.exit(1);}
}
else
  try{   pif=new BufferedReader(
                    new FileReader(
                    filename));
  }catch(Exception e){System.err.println("ERROR: File "+filename+" not found for reading, exiting!"); System.exit(1);}
minConcs1.add(1.0);
minConcs2.add(1.0);
maxConcs1.add(.0);
maxConcs2.add(.0);
return readBufferedPIF(pif,timer);
}

public static Pixel[][][] readBufferedPIF(BufferedReader pif, int timer){
String read ="";
Pixel[][][] ret = new Pixel[sizeX][sizeY][sizeZ];
for(int x=0;x<sizeX;x++)
	for(int y=0;y<sizeY;y++)
		for(int z=0;z<sizeZ;z++)
			ret[x][y][z]=new Pixel();

int x1;
int x2;
int y1;
int y2;
int z1;
int z2;
int id;
try{
read=pif.readLine();
while(read!=null){
//8	NonCondensing	12	12	19	19	0	0	
	String[] parts=read.split("\t");
	//System.out.println(read+" becomes: "+parts[1]+" "+parts[2]+" "+parts[3]+" "+parts[4]+" "+parts[5]+" "+parts[6]+" "+parts[7]);
	if(parts.length>=7){
		id=Integer.parseInt(parts[0]);
		x1=Integer.parseInt(parts[2]);
		x2=Integer.parseInt(parts[3]);
		y1=Integer.parseInt(parts[4]);
		y2=Integer.parseInt(parts[5]);
		z1=Integer.parseInt(parts[6]);
		z2=Integer.parseInt(parts[7]);
		for(int x=x1;x<=x2;x++)
			for(int y=y1;y<=y2;y++)
				for(int z=z1;z<=z2;z++){
					try{
						ret[x][y][z].conc1=Double.parseDouble(parts[8]);
						ret[x][y][z].conc2=Double.parseDouble(parts[9]);
						if(ret[x][y][z].conc1<minConcs1.get(timer))
							minConcs1.set(timer,ret[x][y][z].conc1);
						if(ret[x][y][z].conc2<minConcs2.get(timer))
							minConcs2.set(timer,ret[x][y][z].conc2);
						if(ret[x][y][z].conc1>maxConcs1.get(timer))
							maxConcs1.set(timer,ret[x][y][z].conc1);
						if(ret[x][y][z].conc2>maxConcs2.get(timer))
							maxConcs2.set(timer,ret[x][y][z].conc2);
					}catch(Exception e){}
					ret[x][y][z].type=mapType2int(parts[1]);
					ret[x][y][z].id=id;
					//System.out.println(x+" "+y+" "+z+" = "+parts[1]);
				}
		}
	read=pif.readLine();
	}
}
catch(Exception e){System.err.println("FILE ERROR in readBufferedPIF() "+e);}
return ret;
}

public static Vector<String> dirlist=new Vector<String>();
public static Vector<Pixel[][][]> gridBuffer;
public Pixel[][][] gridPixels;
public int currentPixels;
public boolean loadState=true; //we want to display a grid which hasn't been loaded, display wait message
public boolean loadingFinished=false;//all files available have been read, no more calls to readPIF please

public boolean setGridPixels(int num){
currentPixels=num;
if((num>=0)&&(num<gridBuffer.size())){
  loadState=false;
  gridPixels=new Pixel[gridBuffer.get(num).length][gridBuffer.get(num)[0].length][gridBuffer.get(num)[0][0].length];
	for(int x2=0;x2<sizeX;x2++)
		for(int y2=0;y2<sizeY;y2++)
			gridPixels[x2][y2][whichZ]=gridBuffer.get(num)[x2][y2][whichZ].clone();

	if(trackerID>=0)
	for(int x2=0;x2<sizeX;x2++)
		for(int y2=0;y2<sizeY;y2++)
			if(trackerID==gridPixels[x2][y2][whichZ].id)
				gridPixels[x2][y2][whichZ].type=-1;
  }
else
  loadState=true;
repaint(100);
if(loadState)
  return false;
else
  return true;
}

public grid2D(){
this.addMouseListener(new MouseListener(){
  public void mousePressed(MouseEvent e){
    onClick(e.getX(),e.getY());  //call this method on clicks with mouseposition
  }
  public void mouseEntered(MouseEvent e){}
  public void mouseExited(MouseEvent e){}
  public void mouseClicked(MouseEvent e){}
  public void mouseReleased(MouseEvent e){}
}
);

}

public Dimension getMinimumSize(){
return new Dimension(width,height);
}

public Dimension getMaximumSize(){
return new Dimension(maxWide,maxHigh);
}

public Dimension getPreferredSize(){
return getMaximumSize();
}

public void step(){
  repaint();
}

public int trackerID=-1;

public void onClick(int x,int y){
if(loadState) return;
psXresolution=(int)((getWidth()-15.0)/(sizeX*1.0));
psYresolution=(int)((getHeight()-10.0)/(sizeY*1.0));
int xTrans=x/psXresolution-1;//why???
int yTrans=y/psYresolution;  //why???
if((xTrans<sizeX)&&(yTrans<sizeY)&&(xTrans>0)&&(yTrans>0)){
	trackerID=gridPixels[xTrans][yTrans][whichZ].id;
	setGridPixels(currentPixels);
	}
}

public static int mapType2int(String type){
if((type==null)||(type.equals(""))||(type.equals("Medium")))
  return 0;
if(type.equals("Blue"))
  return 1;
if(type.equals("Red"))
  return 2;
if(type.equals("White"))
  return 3;
return 0;
}

public static void mapType2Col(Graphics g,int celltype){
g.setColor(Color.BLACK);
if(celltype==1)
  g.setColor(Color.BLUE);
if(celltype==2)
  g.setColor(Color.RED);
if(celltype==3)
  g.setColor(Color.WHITE);
if(celltype==0)
  g.setColor(Color.BLACK);
}

static void mapConc2Col(Graphics g,double conc, double min, double max){
if(conc<0.00001){
  g.setColor(Color.BLACK);
  return;
  }//(((conc-min)/(max-min)))
float val=0.7f-0.7f*(float)(((conc-min)/(max-min)));
g.setColor(Color.getHSBColor(val,(float)0.8,(float)1.0));
}

private Image offScreenImage = null;
private Graphics offScreenGraphics;

public synchronized void update( Graphics theG )
{
Dimension d = getSize();
if((offScreenImage == null) ) {
  offScreenImage = createImage( d.width, d.height );
  offScreenGraphics = offScreenImage.getGraphics();
}
paint(offScreenGraphics);
theG.drawImage( offScreenImage, 0, 0, null );
}

public void paint(Graphics g){
if(gridPixels!=null){
int psX=psXstart;
int psY=psYstart;
psXresolution=(int)((getWidth()-15.0)/(sizeX*1.0));
psYresolution=(int)((getHeight()-10.0)/(sizeY*1.0));
  for(int x=0;x<sizeX;x++){
	psY=psYstart;
	psX+=psXresolution;
	for(int y=0;y<sizeY;y++){
		if(gridPixels[x][y][whichZ].type==-1)
			g.setColor(trackerColor);
		else{
			if(mode==0)
				mapType2Col(g,gridPixels[x][y][whichZ].type);
			if(mode==1)
				mapConc2Col(g,gridPixels[x][y][whichZ].conc1,minConcs1.get(currentPixels),maxConcs1.get(currentPixels));
			if(mode==2)
				mapConc2Col(g,gridPixels[x][y][whichZ].conc2,minConcs2.get(currentPixels),maxConcs2.get(currentPixels));
			}
		g.fillRect(psX,psY,psXresolution,psYresolution);
/*		out.println("n "+psX+" "+psY+" m "+(psX+psXresolution)+" "+psY+" l "+(psX+psXresolution)+" "+(psY+psYresolution)+" l "+psX+" "+(psY+psYresolution)+" l");
		if(mode==0)
			out.println(" cp gs "+mapType2Col(gridPixels[x][y][whichZ].type)+" 1.00 shd ef gr ");//add this for black border: gs col0 s gr 
		if(mode==1)
			out.println(" cp gs "+mapConc2Col(gridPixels[x][y][whichZ].conc1,minConc1,maxConc1)+" 1.00 shd ef gr ");//gs col0 s gr 
		if(mode==2)
			out.println(" cp gs "+mapConc2Col(gridPixels[x][y][whichZ].conc2,minConc2,maxConc2)+" 1.00 shd ef gr ");//gs col0 s gr 
		out.println("45.000 slw");
*/		g.setColor(Color.YELLOW);
		//line top
		if(((y>0)&&(gridPixels[x][y][whichZ].id!=gridPixels[x][y-1][whichZ].id))||((y==0)&&(gridPixels[x][y][whichZ].id>=0)))
			g.drawLine(psX,psY,psX+psXresolution,psY);
		//line bottom
		if(((y<sizeY-1)&&(gridPixels[x][y][whichZ].id!=gridPixels[x][y+1][whichZ].id))||((y==sizeY-1)&&(gridPixels[x][y][whichZ].id>=0)))
			g.drawLine(psX,psY+psYresolution,psX+psXresolution,psY+psYresolution);
		//line left
		if(((x>0)&&(gridPixels[x][y][whichZ].id!=gridPixels[x-1][y][whichZ].id))||((x==0)&&(gridPixels[x][y][whichZ].id>=0)))
			g.drawLine(psX,psY,psX,psY+psYresolution);
		//line right
		if(((x<sizeX-1)&&(gridPixels[x][y][whichZ].id!=gridPixels[x+1][y][whichZ].id))||((x==sizeX-1)&&(gridPixels[x][y][whichZ].id>=0)))
			g.drawLine(psX+psXresolution,psY,psX+psXresolution,psY+psYresolution);

		psY+=psYresolution;
	}
  }
 }//end gridpixel!=null
g.setColor(trackerColor);
if(loadState){
  Font t=g.getFont();
  t=new Font(t.getFontName(),Font.BOLD+Font.ITALIC,18);
  g.setFont(t);
  g.drawString("Still loading, please wait...",20,getHeight()/2);
  }
}

}//end grid2D class

public class PIFplayer extends Applet{

public static int maxTrack=100;

private boolean inBrowser=true;
public  grid2D grid;
public controlGrid conti;
public grid track;
public File baseFile=null;

public String getAppletInfo()
{
return "PIFplayer - Author: Johannes Knabe, jknabe@panmental.de - 2008 - Source available at http://panmental.de";
}

public void init(boolean inBrowser,String dir){
this.inBrowser=inBrowser;
this.baseFile=new File(dir);
init();
}

public void init() {
    GridBagLayout gridbag = new GridBagLayout();

    setLayout(gridbag);
    grid=new grid2D();

    GridBagConstraints c1 = new GridBagConstraints();
    c1.anchor=GridBagConstraints.NORTHWEST;
    c1.ipadx=10;
    c1.ipady=0;
    c1.gridx=0;
    c1.gridy=0;
    gridbag.setConstraints(grid, c1);
    add(grid);

   if(inBrowser){
    URL screens=null;
    try{
      screens=new URL(getCodeBase()+getParameter("datafile"));
    }catch(Exception e){System.err.println("ERROR: Invalid URL, exiting! "+getCodeBase()); System.exit(1); }
    conti =new controlGrid(grid,screens);
   }else
    if(baseFile.getName().endsWith(".zip")){
      URL screens=null;
      String url="file://"+baseFile.getAbsolutePath();
      try{
       screens=new URL(url);
      }catch(Exception e){System.err.println("ERROR: Zip file not found, exiting! "+e); System.exit(1); }
      baseFile=null;
      conti =new controlGrid(grid,screens);
     }
     else
      conti =new controlGrid(grid,baseFile);
    GridBagConstraints c3 = new GridBagConstraints();
    c3.anchor=GridBagConstraints.EAST;
    c3.gridx=1;
    c3.gridy=0;
    gridbag.setConstraints(conti, c3);
    add(conti);
}

public void start(){
//int i=10/(1-1);
  if(inBrowser){
    //autostart...
    conti.controlClick();
  }
}

public void stop(){
this.grid=null;
}

public static void main(String[] args){
if((args==null)||(args.length<1)){
  System.err.println("Argument specifying location of .pif files expected, trying current directory...");
  args=new String[1];
  args[0]=".";
}
new AppletFrame("PIFplayer, Graphical Interface",new PIFplayer(),640,480,args[0]);
}
}//end class PIFplayer

class controlGrid extends Container{

public int time=0;
public Label timeLabel;
public Button control;
public Button step;
public Button backstep;
public Choice displayMode;
public javax.swing.Timer actionTimer;
public javax.swing.Timer loadTimer;
public Scrollbar bar;
public static final int width=120; //fix width
public static final int maxhigh=300;   //and max height
public int mode=0;
public final int initDelay=250;
public grid2D grid;
public URL baseDir;
public File baseFile;
public lazyReader background;

public Dimension getMinimumSize(){
return new Dimension(width,maxhigh);
}

public Dimension getMaximumSize(){
return new Dimension(width,maxhigh);
}

public Dimension getPreferredSize(){
return getMaximumSize();
}

public String getStateText(){
 if(mode==2)
  mode=0;
 switch(mode){
  case 0: if(time>0)
            backstep.setEnabled(true);
          else
            backstep.setEnabled(false);
          if(time<grid.dirlist.size())
            step.setEnabled(true); 
          else
            step.setEnabled(false); 
          return "play";
  case 1: step.setEnabled(false); backstep.setEnabled(false); return "pause";
 }
 return "Error";
}

public boolean updateState(){
getStateText();
/*if((time<maxtime)&&(grid.gridBuffer[time]==null))
  grid.gridBuffer[time]=grid.readPIF(baseDir+File.separator+dirlist.get(time),time);
if((time+1<maxtime)&&(grid.gridBuffer[time+1]==null))
  grid.gridBuffer[time+1]=grid.readPIF(baseDir+File.separator+dirlist[time+1],time+1);*/
return grid.setGridPixels(time);
}

public void controlClick(){
 mode++;
 control.setLabel(getStateText());
 if(mode==1)
  actionTimer.start();
 if(mode==0)
  actionTimer.stop();
}

public int memoWhat=0;

public void loadTimerEvent(){
  doStep(memoWhat);
  if(!grid.loadState){
    memoWhat=0;
    loadTimer.stop();
   }
}

public void timerEvent(){
  if(time<grid.dirlist.size()-1){
   if(mode==1)
    doStep(+1);
   if(mode==3)
    doStep(-1);
    }
  else
   if(grid.loadingFinished)
    controlClick();
}

public void doStep(int what){
 loadTimer.stop();
 memoWhat=0;
 grid.loadState=false;
 if((time+what>=grid.dirlist.size())||(time+what<0))
   return;
 time+=what;
 if(updateState()){
  timeLabel.setText(grid.dirlist.get(time).substring(0,grid.dirlist.get(time).length()-4));//filename without ".pif"
 }else{
  time-=what;
  memoWhat=what;
  loadTimer.start();
  }
}

public void stepClick(int what){
 doStep(what);
}

public void delayChanged(){
 actionTimer.setDelay(initDelay-bar.getValue());
}

public void modeChanged(){
    grid.mode=displayMode.getSelectedIndex();
    grid.repaint(100);
}

public void loadPIFfiles(File dir){
String[] ttt=null;
 try{
 File here=dir;
 ttt=here.list(new FilenameFilter(){
                            public boolean accept(File dir, String name){
                              if(name.endsWith("pif"))
                                return true;
                              else
                                return false;
                              }
                            });
try{
 java.util.Arrays.sort(ttt);
}catch(Exception e){System.err.println("ERROR: No .pif files found, exiting."); System.exit(1);}
 for(int i=0;i<ttt.length;i++){
//    System.err.println(ttt[i]);
   grid.dirlist.add(ttt[i]);
 }

}catch(Exception e){System.err.println("Error reading directory! "+e);}

}

public void loadPIFfiles(URL dir){
grid.dirlist=new Vector<String>();
if(dir.toString().endsWith(".zip")){
 grid.screensURL=dir;
 grid.resetScreens();
}else 
{System.err.println("Expected .zip file, got: "+dir);}
}

public void initPIFfiles(){
grid.gridBuffer=new Vector<Pixel[][][]>(grid.dirlist.size());
}

public controlGrid(grid2D grid, URL base){
this.grid=grid;
//baseDir=new File("screens.zip");
baseDir=base;//new File("screens.zip");
loadPIFfiles(baseDir);
background=new lazyReader(baseDir,grid);
initial();
}
public controlGrid(grid2D grid, File base){
this.grid=grid;
baseFile=base;//new File("screens.zip");
loadPIFfiles(baseFile);
background=new lazyReader(baseFile,grid);
initial();
}

public void initial(){
initPIFfiles();
//add control buttons and events
actionTimer=new javax.swing.Timer(initDelay,new ActionListener(){
   public void actionPerformed(ActionEvent e){
    timerEvent();
    }
   }
 );
loadTimer=new javax.swing.Timer(250,new ActionListener(){
   public void actionPerformed(ActionEvent e){
    loadTimerEvent();
    }
   }
 );
setLayout(new GridLayout(0,1));
Label loadprogress=new Label("loaded: 0");
background.loadprogress=loadprogress;
background.start();

add(loadprogress);
add(new Label("time:"));
timeLabel=new Label("0");
add(timeLabel);
 displayMode = new Choice();
 displayMode.add("cell type");
 displayMode.add("morphogen 1");
 displayMode.add("morphogen 2");
displayMode.addItemListener(new ItemListener(){
  public void itemStateChanged(ItemEvent e){
    modeChanged();
  }
}
);
add(displayMode);
add(new Label("speed:"));
bar= new Scrollbar(Scrollbar.HORIZONTAL, initDelay/2-10,20,0, initDelay);//SpeedScrollbar
add(bar);                          //add it 
add(new Label("controls:"));
bar.addAdjustmentListener(new AdjustmentListener(){
  public void adjustmentValueChanged(AdjustmentEvent e){
    delayChanged(); //function called when bar is changed
    }
  }
  );
backstep=new Button("< step");
backstep.addActionListener(new ActionListener(){
  public void actionPerformed(ActionEvent e){
    stepClick(-1);  //call this function when button clicked
    }
  }
  );
step=new Button("step >");
step.addActionListener(new ActionListener(){
  public void actionPerformed(ActionEvent e){
    stepClick(+1);  //call this function when button clicked
    }
  }
  );
control=new Button(getStateText());
add(control);
add(backstep);
add(step);
control.addActionListener(new ActionListener(){
  public void actionPerformed(ActionEvent e){
    controlClick();  //call this function when button clicked
    }
  }
  );
doStep(0);
}

}//end class

class AppletFrame extends JFrame {

private PIFplayer display;

AppletFrame(String title, PIFplayer applet, int width, int height,String dir){
  super(title);
  display=applet;
  applet.init(false,dir);
  this.getContentPane().add(applet);
  this.setSize(width,height);
  this.pack();
  applet.setVisible(true);
  this.setVisible(true);
  addWindowListener(
      new WindowAdapter() {  // close the window when the user clicks its close box
          public void windowClosing(WindowEvent evt) {
          close();
          }
      });
}

public void close(){
   System.exit(0);
   }

}